import { zeroPadValue } from 'ethers';
import { InputMap } from '@noir-lang/noir_js';
import fs from "fs"
import json2toml from "json2toml"
import { execSync } from 'child_process';
import { Verifier, rollup_transaction } from '../circuits/helpers/codegen/rollup_transaction';
import { publish_batch } from '../circuits/helpers/codegen/publish_batch';
import { tx_as_hash } from '../circuits/helpers/codegen/tx_as_hash';
import { Peering } from './peering';
import * as types from './types';


export class BatchBuilder extends Peering {
  owner: types.EthersSigner | undefined;
  anonRPC: string | undefined;

  constructor(_config: types.GlobalConfig) {
    if (typeof global !== 'object') throw new Error("BatchBuilder is not available in the browser.");
    super(_config);
  }

  async initializeBatchBuilder(callback: Function): Promise<void> {
    // // Uncomment below to recompile recursion circuit
    // execSync(`cd ./momiji-helpers/circuits/recursion && nargo compile --only-acir --silence-warnings && ${this.config.nargo} write_vk --silence-warnings && ${this.config.nargo} vk_as_fields --silence-warnings`);
    await this.initializePeering(callback);
  }

  async getHistoricPath(root: string): Promise<string[]> {
    if (root === types.ZERO_VALUE) return types.default_historic_path
    const path: string[] = await this.contracts.state.queryFilter(
      this.contracts.state.filters.BatchPublish(undefined, undefined, root, undefined, undefined))
      .then((event: any) => ((event.length !== 0) ? event[0].args[4] : Promise.reject("No historic path for this root")))
    return path
  }

  async wrapTransaction(transaction: types.Transaction, txProofArtifacts: types.ProofArtifacts): Promise<Verifier> {
    const public_inputs_hash: string = await tx_as_hash(transaction.public_inputs)
    const wrapperInputs: InputMap = {
      public_inputs_hash: public_inputs_hash,
      transaction_verifier: {
        key_hash: txProofArtifacts.vkHash,
        verification_key: txProofArtifacts.vkAsFields,
        proof: txProofArtifacts.proofAsFields,
      }
    }
    let tx_aggregation_object: string[]
    let wrapperArtifacts: types.ProofArtifacts;
    let wrapper_proof = await this.noirs.transaction_wrapper.execute(wrapperInputs)
      .then((wrapper_witness: any) => this.backends.transaction_wrapper.generateProof(wrapper_witness.witness));

    tx_aggregation_object = wrapper_proof.publicInputs.slice(1);
    let artifacts = await this.backends.transaction_wrapper.generateRecursiveProofArtifacts(wrapper_proof, 17);
    wrapperArtifacts = {
      proofData: wrapper_proof,
      proofAsFields: artifacts.proofAsFields,
      vkAsFields: artifacts.vkAsFields,
      vkHash: artifacts.vkHash
    }

    const txWrapperVerifier: Verifier = {
      key_hash: wrapperArtifacts.vkHash,
      proof: wrapperArtifacts.proofAsFields,
      verification_key: wrapperArtifacts.vkAsFields,
      aggregation_object: tx_aggregation_object
    }
    txWrapperVerifier.proof = tx_aggregation_object.concat(txWrapperVerifier.proof)
    return txWrapperVerifier
  }

  async recursiveProof(toml: string, transaction: types.Transaction): Promise<types.RecursionInputs> {
    fs.writeFileSync('./momiji-helpers/circuits/recursion/Prover.toml', toml)
    execSync(`cd ./momiji-helpers/circuits/recursion && nargo execute witness --silence-warnings && ${this.config.nargo} prove --silence-warnings && ${this.config.nargo} proof_as_fields --silence-warnings`);
    const proofAsFields = JSON.parse(fs.readFileSync('./momiji-helpers/circuits/recursion/proofs/proof_fields.json', 'utf8'));
    const vkAsFields = JSON.parse(fs.readFileSync('./momiji-helpers/circuits/recursion/target/vk_fields.json', 'utf8'));
    const rollup: types.RecursionInputs = {
      transaction: transaction,
      accumulator: proofAsFields[0], 
      recursion_verifier: {
        proof: proofAsFields.slice(1),
        verification_key: vkAsFields.slice(1), 
        key_hash: vkAsFields[0],
        aggregation_object: proofAsFields.slice(1, 17)
      }
    }
    return rollup
  }

  async rollupTransaction(
    currentRollup: types.RecursionInputs[],
    transaction: types.Transaction
  ): Promise<types.RecursionInputs[]> {

    const tx_verifier = {
      key_hash: transaction.proof_artifacts.vkHash,
      verification_key: transaction.proof_artifacts.vkAsFields,
      proof: transaction.proof_artifacts.proofAsFields
    }

    const idx: number = currentRollup.length > 0 ? currentRollup.length - 1 : 0;
    const recursion_verifier = (currentRollup.length === 0) ? await this.wrapTransaction(transaction, transaction.proof_artifacts) : currentRollup[idx].recursion_verifier;
    const previous_accumulator: string = (currentRollup.length === 0) ? types.ZERO_VALUE : currentRollup[idx].accumulator
    
    const toml: string = await rollup_transaction(
      tx_verifier,
      recursion_verifier!,
      previous_accumulator,
      transaction.public_inputs
    ).then((rollupInputs: any) => json2toml(rollupInputs))

    const newTx = await this.recursiveProof(toml, transaction)
    currentRollup.push(newTx);
    return currentRollup
  }

  async preparePublish(rollup: types.RecursionInputs[]): Promise<types.ContractPublish> {
    if (rollup.length == 0) return Promise.reject("❌ Empty rollup cannot be published");
    const finalRollup: types.RecursionInputs = rollup[rollup.length - 1];
    const validRoots: string[] = await this.contracts.state.getValidRoots()
    const merkleRoot: string = await this.contracts.state.merkleRoot()
    const index: number = validRoots.indexOf(merkleRoot)
    const txVerifier: types.ProofArtifacts = finalRollup.transaction!.proof_artifacts;
    const publisherHelperInputs: types.PublishInputs = {
      accumulator: finalRollup.accumulator,
      hist_tree_input: {
        root: await this.contracts.state.histRoot(),
        leaf: merkleRoot,
        index: types.toFixedHex(index, true),
        path: await this.getHistoricPath(merkleRoot),
        utxo_roots: new Array(16).fill("0x").map((_, i, __) => {
          if (rollup[i] !== undefined && rollup[i].transaction !== undefined) return rollup[i]!.transaction!.public_inputs.utxo_root;
          else return types.ZERO_VALUE;
        })
      },
      tx_verifier: {
        proof: txVerifier.proofAsFields,
        verification_key: txVerifier.vkAsFields,
        key_hash: txVerifier.vkHash
      },
      recursion_verifier: finalRollup.recursion_verifier!
    }

    if (this.verbose) fs.writeFileSync("./test_files/publisherHelperInputs.json", JSON.stringify(publisherHelperInputs))

    const formattedPublishInputs = await publish_batch(
      publisherHelperInputs.accumulator,
      publisherHelperInputs.hist_tree_input,
      publisherHelperInputs.tx_verifier,
      publisherHelperInputs.recursion_verifier
    )
    if (this.verbose) fs.writeFileSync("./test_files/formattedPublishInputs.json", JSON.stringify(formattedPublishInputs))
    const toml = json2toml(formattedPublishInputs as InputMap)
    fs.writeFileSync('./momiji-helpers/circuits/publish/Prover.toml', toml)
    execSync(`cd ./momiji-helpers/circuits/publish && nargo prove --silence-warnings && nargo verify --silence-warnings`);
    const verifierInputs = {
      proof: fs.readFileSync('./momiji-helpers/circuits/publish/proofs/publish.proof', 'utf-8'),
      pi: formattedPublishInputs.pi_contract_hash
    }

    const batch_publish: types.ContractBatch = {
      tx_key_hash: zeroPadValue(formattedPublishInputs.tx_verifier.key_hash, 32),
      recursive_key_hash: zeroPadValue(formattedPublishInputs.recursion_verifier.key_hash, 32),
      new_root: zeroPadValue(formattedPublishInputs.batch.new_state_root, 32),
      old_hist_root: zeroPadValue(formattedPublishInputs.batch.hist_tree.root, 32),
      new_hist_root: zeroPadValue(formattedPublishInputs.batch.hist_tree.new_root, 32),
      oracle: zeroPadValue(formattedPublishInputs.batch.batch_oracle, 32),
      historic_path: formattedPublishInputs.batch.hist_tree.new_path.map((p: any) => zeroPadValue(p, 32)),
      aggregation_object: new Array(16).fill("0x")
        .map((_, i, __) => "0x" + verifierInputs.proof.slice(i * 64, (i + 1) * 64)),
      transactions: rollup.map(r => r.transaction).map(tx => tx!.contract_inputs)
    }
    if (this.verbose) fs.writeFileSync("./test_files/batch_publish_deposit.json", JSON.stringify(batch_publish))
    return {
      proof: "0x" + verifierInputs.proof.slice(1024),
      batch: batch_publish
    } as types.ContractPublish
  }

}