import { Provider, Signer, connectWallet } from "./connect-wallet";
import { Contract, formatEther, parseEther, AbiCoder, ParamType } from "ethers";
import { addresses } from "./constants";
import erc20ABI from "../abis/ERC20.json";
import stateABI from "../abis/TechnicalPreview.json";
import tokenSwapABI from "../abis/TokenSwap.json";
import UniswapStakerABI from "../abis/UniswapV3Staker.json";
import LiqMinerABI from "../abis/Miner.json";
import unlockedLiqMinerABI from "../abis/UnlockedMiner.json";
import UniswapLiqPool from "../abis/UniswapV3Pool.json";
import NonFungibleABI from "../abis/NonfungiblePositionManager.json";
import technicalPreviewABI from "../abis/TechnicalPreview.json"
import verifierABI from "../abis/IVerifier.json"
import { getUtxoBalance } from "./zkhelpers";
// Define a type for the slice state
export interface Balance {
    address: string;
    ETH: string | number;
    oldXFT: string | number;
    XFT: string | number;
    anonUSD: string | number;
    anonBTC: string | number;
    zkXFT: string | number;
}

export interface PriceFeed {
    XFT: string;
    ETH: string;
    anonUSD: string
    anonBTC: string
}

export const getContracts = async () => {
    const contracts = {
        state: new Contract(addresses.technicalPreview, stateABI, Signer),
        XFT: new Contract(addresses.XFT_LATEST, erc20ABI, Provider),
        verifier: new Contract(addresses.verifier, verifierABI, Signer)
    };
    return contracts;
}

// TODO: Not this -- don't reinit contracts every time they are called vvv

export const getSelectedAddress = async () => {
    const accounts = await Provider.send("eth_requestAccounts", []);
    return accounts[0];
};

export const getEthBalance = async () => {
    const address = await getSelectedAddress();
    const balance = await Provider.getBalance(address);
    return balance;
};

export const getERC20Balance = async (tokenAddress: string) => {
    const address = await getSelectedAddress();
    const contract = new Contract(tokenAddress, erc20ABI, Provider);
    const balance = await contract.balanceOf(address);
    return balance;
};

export const getOldXFTBalance = async () => {
    const balance = await getERC20Balance(addresses.XFT_OLD);
    return balance;
};

export const getXFTBalance = async () => {
    const balance = await getERC20Balance(addresses.XFT_LATEST);
    return balance;
};

export const approveToken = async (tokenAddress: string, spender: string, amount: string | bigint) => {
    const tokenContract = new Contract(tokenAddress, erc20ABI, Signer)
    await tokenContract.approve(spender, amount)
}

export const getAll = async () => {
    const b: Balance = {
        address: await getSelectedAddress(),
        ETH: roundTo4Decimals(formatEther(await getEthBalance())),
        oldXFT: roundTo4Decimals(formatEther(await getOldXFTBalance())),
        XFT: roundTo4Decimals(formatEther(await getXFTBalance())),
        anonUSD: roundTo4Decimals(
            formatEther(await getERC20Balance(addresses.anonUSD))
        ),
        anonBTC: roundTo4Decimals(
            formatEther(await getERC20Balance(addresses.anonBTC))
        ),
        zkXFT: roundTo4Decimals(formatEther(await getUtxoBalance()))
    };
    return b;
};

function roundTo4Decimals(number: string) {
    const rounded = parseFloat(number).toFixed(4);
    return Number(rounded);
}

export const upgrade = async () => {
    const tokenUpgrade = new Contract(
        "0x010a526C562F14ECC97A300aDa11E38ab6D985a5",
        tokenSwapABI,
        Signer
    );
    const oldXFTSign = new Contract(addresses.XFT_OLD, erc20ABI, Signer);
    const oldXFT = new Contract(addresses.XFT_OLD, erc20ABI, Provider);
    if (
        (await oldXFT.allowance(
            await getSelectedAddress(),
            await tokenUpgrade.getAddress()
        )) == "0"
    ) {
        await oldXFTSign.approve(
            await tokenUpgrade.getAddress(),
            await oldXFT.totalSupply()
        );
    }
    await tokenUpgrade.upgrade();
};
export const getXFTEther = async() => {
    const UniswapLiqContract = new Contract("0x701B319AFF36311F720f0a1ebD6636Dd84Bb3E8B", UniswapLiqPool, Provider);
    const sqrtPriceX96: bigint = await UniswapLiqContract.slot0().then(slot0 => slot0.sqrtPriceX96)
    const xftToEther = formatEther(BigInt(1e18) * sqrtPriceX96 * sqrtPriceX96 / BigInt(2**192))
    return parseFloat(xftToEther) 
    
}
export const manualStake = async (tokenId: string) => {
    const MinerContract = new Contract(addresses.LIQ_MINER_UNLOCKED, LiqMinerABI, Provider)
    const NonfungiblePositionManager = new Contract(addresses.NonFungiblePositionManager, NonFungibleABI, Signer)
    const incentiveKey = await MinerContract.incentiveKey();
    const keyFormatted = incentiveKey.map((r: any) => {
        return r
    })
    const coder = AbiCoder.defaultAbiCoder()
    await NonfungiblePositionManager["safeTransferFrom(address,address,uint256,bytes)"](
        await getSelectedAddress(),
        addresses.UNIV3_STAKER,
        tokenId,
        coder.encode(["tuple(address rewardToken, address pool, uint256 startTime, uint256 endTime, address refundee)"], [keyFormatted]),
    )
};

export const manualWithdraw = async (tokenId: string) => {
    const V3StakerContract = new Contract(addresses.UNIV3_STAKER, UniswapStakerABI, Signer)
    const MinerContract = new Contract(addresses.LIQ_MINER_UNLOCKED, LiqMinerABI, Provider)
    const incentiveKey = await MinerContract.incentiveKey()
    const keyFormatted = incentiveKey.map((r: any) => {
        return r
    })
    await V3StakerContract.unstakeToken(keyFormatted, tokenId)
    await V3StakerContract.withdrawToken(tokenId, await getSelectedAddress(), "0x0000000000000000000000000000000000000000000000000000000000000000")
}

export const autoStake = async (amount1: string, amount2: string) => {
    const MinerContract = new Contract(addresses.LIQ_MINER_UNLOCKED, unlockedLiqMinerABI, Signer)
    const value1 = parseEther(amount1)
    const value2 = parseEther(amount2)
    await approveToken(addresses.anonUSD, addresses.LIQ_MINER_UNLOCKED, value1)
    await MinerContract.addLiquidity(value1, { value: value2 })
}

export const autoUnstake = async (id: string) => {
    const MinerContract = new Contract(addresses.LIQ_MINER_UNLOCKED, LiqMinerABI, Signer)
    await MinerContract.withdraw(id);
}

export const getRewardInfo = async (tokenIDs: string[]) => {
    const V3StakerContract = new Contract(addresses.UNIV3_STAKER, UniswapStakerABI, Signer)
    const MinerContract = new Contract(addresses.LIQ_MINER_UNLOCKED, LiqMinerABI, Provider)
    const incentiveKey = await MinerContract.incentiveKey()
    const keyFormatted = incentiveKey.map((r: any) => {
        return r
    })
    let rewardInfo: any[] = []
    let total = 0

    for (let i = 0; i < tokenIDs.length; i++) {
        rewardInfo.push(await V3StakerContract.getRewardInfo.staticCall(
            keyFormatted,
            tokenIDs[i]
        ))
        total += rewardInfo[i][0]
    }
    return total
}

export const requiredStakingETH = async (amount: string) => {
    const MinerContract = new Contract(addresses.LIQ_MINER_UNLOCKED, LiqMinerABI, Provider)
    const parsedAmount = parseEther(amount)
    return formatEther(await MinerContract.getETHAmount(parsedAmount))
}

export const getAutoStakes = async () => {
    const MinerContract = new Contract(addresses.LIQ_MINER_UNLOCKED, LiqMinerABI, Provider)
    const V3StakerContract = new Contract(addresses.UNIV3_STAKER, UniswapStakerABI, Provider)
    const StakeFilter = MinerContract.filters.NFT_LOCKED(null, await getSelectedAddress())
    const allStakes: any[] = await MinerContract.queryFilter(StakeFilter)

    let stillStaked: string[] = []
    for (let i = 0; i < allStakes.length; i++) {
        if ((await V3StakerContract.deposits(allStakes[i].args[0].toString())).owner == addresses.LIQ_MINER_UNLOCKED) {
            stillStaked.push(allStakes[i].args[0].toString())
        }
    }
    return stillStaked
}

export const getManualStakes = async () => {
    const NonfungiblePositionManager = new Contract(addresses.NonFungiblePositionManager, NonFungibleABI, Provider)
    const TransferFilter = NonfungiblePositionManager.filters.Transfer(await getSelectedAddress(), addresses.UNIV3_STAKER, null)
    const allStakes: any[] = await NonfungiblePositionManager.queryFilter(TransferFilter)
    let stillStaked: string[] = []

    for (let i = 0; i < allStakes.length; i++) {
        if (await NonfungiblePositionManager.ownerOf(allStakes[i].args[2].toString()) == addresses.UNIV3_STAKER) {
            stillStaked.push(allStakes[i].args[2].toString())
        }
    }
    return [...new Set(stillStaked)]
}

export const ClaimRewardsManual = async (tokenID: string) => {
    const V3StakerContract = new Contract(addresses.UNIV3_STAKER, UniswapStakerABI, Signer)
    const MinerContract = new Contract(addresses.LIQ_MINER_UNLOCKED, LiqMinerABI, Provider)
    const incentiveKey = await MinerContract.incentiveKey()
    const keyFormatted = incentiveKey.map((r: any) => {
        return r
    })


    const rewardInfo = await V3StakerContract.getRewardInfo.staticCall(
        keyFormatted,
        tokenID
    )
    await V3StakerContract.claimReward(addresses.XFT_LATEST, await getSelectedAddress(), rewardInfo[0])
}

export const getMerkleRoot = async () => {
    const technicalPreviewContract = new Contract(addresses.technicalPreview, technicalPreviewABI, Provider)
    const merkleRoot = await technicalPreviewContract.merkleRoot()
    return merkleRoot
}

export const getUSD = async () => {
    const options = {method: 'GET', headers: {accept: 'application/json'}};
    const url = `https://min-api.cryptocompare.com/data/pricemultifull?fsyms=ETH&tsyms=USD`
    const res = await fetch(url, options)
    const price = (await res.json())["RAW"]["ETH"]["USD"]["PRICE"]
    console.log(`The price ${price}`)
    return price
}

export const getTotal = async () => {
    let p: PriceFeed = {
        ETH: (await getUSD()).toString(),
        XFT: (await getXFTEther() * await getUSD()).toString(),
        anonUSD: "1.00",
        anonBTC: "1",
    }
    return p
}