import { Provider, Signer } from "./connect-wallet"
import { pki, random } from "node-forge";
import { ethers } from "ethers";
import TechnicalPreview from "../abis/TechnicalPreview.json"
import { hashDB, UTXO_DB } from "./db";
import { addresses } from "./constants";
import { Buffer } from "buffer"
import { GlobalConfig } from "../../momiji-helpers/utils/types";
import { addUTXO, getUtxoStatus } from "./zkhelpers";
import { TransactionBuilder } from "../../momiji-helpers/utils/transactionBuilder";
import { ZERO_VALUE } from "../../momiji-helpers/utils/types"

const buf2hex = (buffer) => { // buffer is an ArrayBuffer
    return [...new Uint8Array(buffer)]
        .map(x => x.toString(16).padStart(2, '0'))
        .join('');
}

export const deriveKeyFromSignature = async (msg: string) => {
    const signature = await Signer.signMessage(msg)
    let enc = new TextEncoder()
    const encodedMessage = enc.encode(signature + signature)
    const hashBuffer = await crypto.subtle.digest("SHA-256", encodedMessage)
    await addHash(buf2hex(hashBuffer))
    const seed = buf2hex(hashBuffer)
    const prng = random.createInstance()
    prng.seedFileSync = () => seed
    const { privateKey, publicKey } = pki.rsa.generateKeyPair({ bits: 4096, prng, workers: 6 })
    const newHash = await crypto.subtle.digest("SHA-256", enc.encode((privateKey.d.add(privateKey.e)).toString()))
    await addHash(buf2hex(newHash))

    return [privateKey, publicKey]

}

export const deriveKeyFromHash = async (hashString) => {
    // const hashBuffer = new Uint8Array(hashString.match(/[\da-f]{2}/gi).map(function (hashString) {
    //     return parseInt(hashString, 16)
    // }))
    const seed = hashString//hashBuffer.buffer
    const prng = random.createInstance()
    prng.seedFileSync = () => seed
    const { privateKey, publicKey } = pki.rsa.generateKeyPair({ bits: 4096, prng, workers: 6 })

    return [privateKey, publicKey]
}

export const deriveKeyFromPrevious = async () => {
    const hashes = await hashDB.hash.toArray()

    const prevHash = hashes.at(-2).hash
    const hash = hashes.at(-1).hash
    let enc = new TextEncoder()
    const encodedMessage = enc.encode(prevHash + hash)
    const hashBuffer = await crypto.subtle.digest("SHA-256", encodedMessage)
    await addHash(buf2hex(hashBuffer))
    const seed = buf2hex(hashBuffer)
    const prng = random.createInstance()
    prng.seedFileSync = () => seed
    const { privateKey, publicKey } = pki.rsa.generateKeyPair({ bits: 4096, prng, workers: 6 })
    return [privateKey, publicKey]
}

export const getSecret = async () => {
    const avaiableSeeds = await hashDB.hash.toArray()
    let [privKey, pubKey] = [undefined, undefined]
    if (avaiableSeeds.length == 0) {
        [privKey, pubKey] = await deriveKeyFromSignature("Momiji Protocol - Wallet Initialization")
    } else {
        [privKey, pubKey] = await deriveKeyFromPrevious()
    }
    return [privKey, pubKey]
}

export const recover = async () => {
    const globalConfig: GlobalConfig = {
        signer: Signer,
        provider: Provider
    }
    const transactionBuilder = new TransactionBuilder(globalConfig)
    const filter = transactionBuilder.contracts.state.filters.TransactionPublish(null, null, null)
    let events = await transactionBuilder.contracts.state.queryFilter(filter)
    let searching = true
    let decryptedNotes = []
    let latest = 0
    let threshold = 0

    while (searching) {
        latest = decryptedNotes.length
        const [privK, pubK] = await getSecret()
        for (let i = 0; i <= events.length; i++) {
            if (events[i] !== undefined) {
                const encryptedNoteArray = events[1].args[0].encrypted_utxo
                .filter(utxo => utxo.secret !== ZERO_VALUE)
                for (let j = 0; j < encryptedNoteArray.length; j++) {
                    try {
                        const EncryptedNote = {
                            secret: Buffer.from(encryptedNoteArray[j].secret.slice(2), "hex"),
                            amount: Buffer.from(encryptedNoteArray[j].amount.slice(2), "hex"),
                            asset_type: Buffer.from(encryptedNoteArray[j].data.slice(2), "hex") 
                        }
                        const decrypted = {
                            secret: decryptNote(privK, EncryptedNote.secret),
                            amount: decryptNote(privK, EncryptedNote.amount),
                            asset_type: decryptNote(privK, EncryptedNote.asset_type)
                        }
                        const commitment: string = await transactionBuilder.utxo_to_commitment(decrypted)
                        decryptedNotes.push({ "keys": [privK, pubK], "note": decrypted })
                        
                        addUTXO({
                            commitment: commitment,
                            secret:decrypted.secret, 
                            amount:decrypted.amount, 
                            assetType: decrypted.asset_type, 
                            timestamp: Math.floor(new Date().getTime() / 1000),
                            status: "queued", 
                            chainStatus: "mined"
                        })
                        events = events.slice(i + 1)
                        break
                    } catch (e) {
                        // console.error(e)
                        console.log("cannot decrypt note")
                    }
                }
                
            }

        }
        if (decryptedNotes.length == latest) {
            threshold += 1
            if (threshold > 3) {
                searching = false
            }
        }

    }
    await getUtxoStatus()
    await getUtxoStatus()

}

export const encryptNote = (publicKey, msg) => {
    return publicKey.encrypt(msg)
}

export const decryptNote = (privateKey, encryptedMessage) => {
    return privateKey.decrypt(encryptedMessage)
}

const addHash = async (hash: string) => {
    try {
        const id = await hashDB.hash.add({ hash })
    }
    catch (e) {
        console.error(e)
    }
}


export const generateLocalEncryption = async (password) => {
    const enc = new TextEncoder()
    const algo = {
        name: "PBKDF2",
        hash: 'SHA-256',
        salt: enc.encode('offshift891265'),
        iterations: 1000
    }
    const Key = await crypto.subtle.deriveKey(
        algo,
        await crypto.subtle.importKey(
            "raw",
            enc.encode(password),
            algo,
            false,
            ["deriveKey"]
        ),
        { name: 'AES-GCM', length: 256 },
        false,
        ["encrypt", "decrypt"]
    )
    const iv = window.crypto.getRandomValues(new Uint8Array(12));
    localStorage.setItem("iv", JSON.stringify(iv))
    return Key
}


export const encryptLocally = async (msg, Key) => {
    const iv = new Uint8Array(JSON.parse(localStorage.getItem("iv")))
    const encryptedMessage = await crypto.subtle.encrypt(
        { name: "AES-GCM", iv },
        Key,
        new TextEncoder().encode(msg)
    )
    return encryptedMessage
}

export const decryptLocally = async (msg, Key) => {
    const iv = new Uint8Array(JSON.parse(localStorage.getItem("iv")))
    const decryptedMessage = await crypto.subtle.decrypt(
        { name: "AES-GCM", iv },
        Key,
        msg
    )
    return decryptedMessage
}

// export 